// NOTE: The word 'slave' appears throughout this file repeatedly. Slave refers to
// a parallel process (often on some machine in a Beowulf cluster) designated for
// the execution of work as opposed to the 'master' which is a process that coordinates
// the progress of the slaves and collects results. Any similarity to a real slave
// either in fact or fiction is purely coincidental

/*

    prand.so - A block parallelization of glibc's random() function
    Copyright (C) 2003  Jason Main (jmain@onyx.boisestate.edu)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
  Parts of this file derive from random.c in the gnu glibc code. This gnu code is
  released under the following copyright.
*/

/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. [rescinded 22 July 1999]
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * This is derived from the Berkeley source:
 * random.c	5.5 (Berkeley) 7/6/88
 * It was reworked for the GNU C Library by Roland McGrath.
 */

#define TRUE		0xFFFFFF
#define FALSE		0

#include <errno.h>
//#include <iostream.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>  // must include stdio  for pvm3.h to compile correctly
#include <sys/times.h> /* for times system call */
#include <sys/time.h>  /* for gettimeofday system call */
#include <stdint.h> 

#include "prand.h"		//header files for the random number system

#if __WORDSIZE == 64
typedef uint32_t		rand_int32_t;
#else
typedef unsigned long		rand_int32_t;
#endif

// Forward declarations. We may or may not need these depending on whether the debug
// flag was set in prand.h
/*extern int degrees[];*/
/*extern int seps[];*/
/*extern long int randtbl[];*/
/*extern long int *fptr;*/
/*extern long int *rptr;*/
/*extern long int *state;*/
/*extern int rand_type;*/
/*extern int rand_deg;*/
/*extern int rand_sep;*/
/*extern long int *end_ptr;*/

extern const rand_int32_t invPowerArray[31];
int exportSlibcState();
int importSlibcState();
void identity(rand_int32_t** a,int n);
int initStaticArray(rand_int32_t*** p,int n);
void deallocStaticArray(rand_int32_t*** p);
rand_int32_t invPowerOfA(rand_int32_t power);
unsigned long fastPower(unsigned long a, unsigned long n);
unsigned long fastPoly(unsigned long a, unsigned long n);
unsigned long fastPolyItr(unsigned long* powers, unsigned long x,unsigned long n);
int log_2(unsigned long n);
void matrixMult(rand_int32_t** left,rand_int32_t** right,
                rand_int32_t** res,int n);
void updateStateMatrix(rand_int32_t ** a, int n);
void createPermMatrix(rand_int32_t** a);
void createPermMatrixRev(rand_int32_t** a);
void iterativeRand(long stride);
long int t_random_backwords();    //Needed for moving through internal 
long int t_random();            //state space

//******************** random.c stuff *************************
// 
// The following lines were imported from random.c and modified to
// allow unrankRand() to work. These functions/declarations represent
// the internal information that unrankRand() must have about the 
// operation of random() in order to work correctly

#if 0

#include <ansidecl.h>
#include <limits.h>
#include <stddef.h>
#include <stdlib.h>

#else

#define	ULONG_MAX  ((unsigned long)(~0L))     /* 0xFFFFFFFF for 32-bits */
#define	LONG_MAX   ((long)(ULONG_MAX >> 1))   /* 0x7FFFFFFF for 32-bits*/

#ifdef __STDC__
#  define PTR void *
#  ifndef NULL
#    define NULL (void *) 0
#  endif
#else
#  define PTR char *
#  ifndef NULL
#    define NULL (void *) 0
#  endif
#endif

#endif

/* For each of the currently supported random number generators, we have a
   break value on the amount of state information (you need at least this many
   bytes of state info to support this random number generator), a degree for
   the polynomial (actually a trinomial) that the R.N.G. is based on, and
   separation between the two lower order coefficients of the trinomial.  */

/* Linear congruential.  */
#define	TYPE_0		0
#define	BREAK_0		8
#define	DEG_0		0
#define	SEP_0		0

/* x**7 + x**3 + 1.  */
#define	TYPE_1		1
#define	BREAK_1		32
#define	DEG_1		7
#define	SEP_1		3

/* x**15 + x + 1.  */
#define	TYPE_2		2
#define	BREAK_2		64
#define	DEG_2		15
#define	SEP_2		1

/* x**31 + x**3 + 1.  */
#define	TYPE_3		3
#define	BREAK_3		128
#define	DEG_3		31
#define	SEP_3		3

/* x**63 + x + 1.  */
#define	TYPE_4		4
#define	BREAK_4		256
#define	DEG_4		63
#define	SEP_4		1


/* Array versions of the above information to make code run faster.
   Relies on fact that TYPE_i == i.  */

#define	MAX_TYPES	5	/* Max number of types above.  */
#define LARGEST_POLY_SIZE   64

RANDOM_DEC int degrees[MAX_TYPES] = { DEG_0, DEG_1, DEG_2, DEG_3, DEG_4 };
RANDOM_DEC int seps[MAX_TYPES] = { SEP_0, SEP_1, SEP_2, SEP_3, SEP_4 };
RANDOM_DEC int breaks[MAX_TYPES] = {BREAK_0,BREAK_1,BREAK_1,BREAK_3,BREAK_4 };
RANDOM_DEC sem_t masterSem;
RANDOM_DEC int masterSem_init = FALSE;
int masterReady = 0;
RANDOM_DEC sem_t unrankGlobalSem;
RANDOM_DEC int unrankGlobalSem_init = FALSE;

/* Initially, everything is set up as if from:
	initstate(1, randtbl, 128);
   Note that this initialization takes advantage of the fact that srandom
   advances the front and rear pointers 10*rand_deg times, and hence the
   rear pointer which starts at 0 will also end up at zero; thus the zeroeth
   element of the state information, which contains info about the current
   position of the rear pointer is just
	(MAX_TYPES * (rptr - state)) + TYPE_3 == TYPE_3.  */

//NOTE: We make this a larger allocation so that we can handle importing type
//      The version in random.c has [DEG_3+1] for the default setting
RANDOM_DEC rand_int32_t randtbl[DEG_4 + 1] =
  { TYPE_3,
    -1726662223, 379960547, 1735697613, 1040273694, 1313901226,
    1627687941, -179304937, -2073333483, 1780058412, -1989503057,
    -615974602, 344556628, 939512070, -1249116260, 1507946756,
    -812545463, 154635395, 1388815473, -1926676823, 525320961,
    -1009028674, 968117788, -123449607, 1284210865, 435012392,
    -2017506339, -911064859, -370259173, 1132637927, 1398500161,
    -205601318,
    };

/* FPTR and RPTR are two pointers into the state info, a front and a rear
   pointer.  These two pointers are always rand_sep places apart, as they
   cycle through the state information.  (Yes, this does mean we could get
   away with just one pointer, but the code for random is more efficient
   this way).  The pointers are left positioned as they would be from the call:
	initstate(1, randtbl, 128);
   (The position of the rear pointer, rptr, is really 0 (as explained above
   in the initialization of randtbl) because the state table pointer is set
   to point to randtbl[1] (as explained below).)  */

RANDOM_DEC rand_int32_t *fptr = &randtbl[SEP_3 + 1];	//Point the end of a TYPE_3 range
RANDOM_DEC rand_int32_t *rptr = &randtbl[1];

/* The following things are the pointer to the state information table,
   the type of the current generator, the degree of the current polynomial
   being used, and the separation between the two pointers.
   Note that for efficiency of random, we remember the first location of
   the state information, not the zeroth.  Hence it is valid to access
   state[-1], which is used to store the type of the R.N.G.
   Also, we remember the last location, since this is more efficient than
   indexing every time to find the address of the last element to see if
   the front and rear pointers have wrapped.  */

RANDOM_DEC rand_int32_t *state = &randtbl[1];

RANDOM_DEC int rand_type = TYPE_3;
RANDOM_DEC int rand_deg = DEG_3;
RANDOM_DEC int rand_sep = SEP_3;

//static long int *end_ptr = &randtbl[sizeof(randtbl) / sizeof(randtbl[0])];
RANDOM_DEC rand_int32_t *end_ptr = &randtbl[SEP_3 + 1];
//NOTE: The declaration has changed a little to set the function in a consistent
//      state for a TYPE_3 poly even though we no have enough space
//      for a TYPE_4 poly. This may be overkill as it's likely to never
//      be used

/* Initialize the random number generator based on the given seed.  If the
   type is the trivial no-state-information type, just remember the seed.
   Otherwise, initializes state[] based on the given "seed" via a linear
   congruent generator.  Then, the pointers are set to known locations
   that are exactly rand_sep places apart.  Lastly, it cycles the state
   information a given number of times to get rid of any initial dependencies
   introduced by the L.C.R.N.G.  Note that the initialization of randtbl[]
   for default usage relies on values produced by this routine.  */
void t_srandom ( unsigned int x)
{
  state[0] = x;
  if (rand_type != TYPE_0)
    {
      register long int i;
      for (i = 1; i < rand_deg; ++i)
	state[i] = (1103515145 * state[i - 1]) + 12345;
      fptr = &state[rand_sep];
      rptr = &state[0];
      for (i = 0; i < 10 * rand_deg; ++i)
	random();
    }
}

/* If we are using the trivial TYPE_0 R.N.G., just do the old linear
   congruent bit.  Otherwise, we do our fancy trinomial stuff, which is the
   same in all the other cases due to all the global variables that have been
   set up.  The basic operation is to add the number at the rear pointer into
   the one at the front pointer.  Then both pointers are advanced to the next
   location cyclically in the table.  The value returned is the sum generated,
   reduced to 31 bits by throwing away the "least random" low bit.
   Note: The code takes advantage of the fact that both the front and
   rear pointers can't wrap on the same call by not testing the rear
   pointer if the front one has wrapped.  Returns a 31-bit random number.  */

long int t_random ()
{
  if (rand_type == TYPE_0)
    {
      state[0] = ((state[0] * 1103515245) + 12345) & LONG_MAX;
      return state[0];
    }
  else
    {
      long int i;
      *fptr += *rptr;
      /* Chucking least random bit.  */
      i = (*fptr >> 1) & LONG_MAX;
      ++fptr;
      if (fptr >= end_ptr)
	{
	  fptr = state;
	  ++rptr;
	}
      else
	{
	  ++rptr;
	  if (rptr >= end_ptr)
	    rptr = state;
	}
      return i;
    }
}

// ************************ Functions to export *********************
#define RAND_A  1103515245    //These are the linear congruent params
#define RAND_B  12345

int unrankRand(long long int stride)
/*
PURPOSE: This function takes the current random() state and permutes it as if stride many calls 
        to random() have already been performed. This function is intended to implement a 
        block based parallization of the standard glibc random function.
		
		Returns true if there was a problem
*/
{
  unsigned long long dist;
  long mod;
  int n;
  rand_int32_t tmpLong;

  rand_int32_t** t1;	//Dynamically allocated temp arrays
  rand_int32_t** t2;	
  rand_int32_t** t3;

  rand_int32_t** s_source;	//Pointers to the temp arrays used to avoid memcpy()s
  rand_int32_t** tmp;
  rand_int32_t** result;
  rand_int32_t** swap;
  int errorCode;

  if(stride == 0) //Need to handle this separately for TYPE_O
      return 0;

  if( unrankGlobalSem_init == FALSE)	//Init the sem
  {
	  errorCode = par_initialize();
	  if(errorCode)
		  return errorCode;
  }
  if(sem_wait(&unrankGlobalSem))
     return UNRANK_FAILED_TO_GET_SEMAPHORE;

  errorCode = importSlibcState();	//Get the state that's being asked to be permuted
  if(errorCode)
  {
	  sem_post(&unrankGlobalSem);		//We have an error so release
	  return errorCode;
  }

  if(rand_type == TYPE_0)   //We are using linear congruence so handle this separately
  //Based off of state[0] = ((state[0] * 1103515245) + 12345) & LONG_MAX;
  {
      if( stride < 0)
      {
          //Compute r_n = r_(n+k)-b( a^(k-1) + a^(k-2) + .... + a + 1 )
          //Where r_(n+k) is the current state[0], k is -stride, (a,b) are the linear congruent
          //parameters, r_n is state prior to stride many steps being taken, and
          //M is (LONG_MAX+1)

          stride = -stride;
          tmpLong = (RAND_B*fastPoly(RAND_A, stride-1)) & LONG_MAX; //Compute b( a^(k-1)+...+ 1 )
          tmpLong = (state[0]-tmpLong) & LONG_MAX;        //Compute r_(n+k)-b( a^(k-1)+...+ 1 )
          state[0] = (tmpLong*invPowerOfA(stride)) & LONG_MAX;
          //Compute r_n = r_(n+k)-b( a^(k-1) + a^(k-2) + .... + a + 1 )
          //Finished the inversion
      }
      else
      {
          //Compute r_(n+k) = (a^k)(r_n) + b( a^(k-1) + a^(k-2) + .... + a + 1 ) Mod M
          //Where r_n is the current state[0], k is stride, (a,b) are the linear congruent
          //parameters, r_(n+k) is state after stride many steps have been taken, and
          //M is (LONG_MAX+1)
          
          tmpLong = fastPower(RAND_A, stride); //compute a^stride
          //r_n+k  =  (a^k   * r_n )   +      b*(  a^(k-1)+.... + a + 1  ) mod M 
          state[0] = (tmpLong*state[0] + RAND_B*fastPoly(RAND_A,stride-1)) & LONG_MAX;
      }

      errorCode = exportSlibcState();	//Export the state thats being asked to be permuted
	  if(sem_post(&unrankGlobalSem))
		  return UNRANK_FAILED_TO_GET_SEMAPHORE;
	  return errorCode;
  }

  //Else we are a lagged fibonacci generator so....

  n=rand_deg;		//Get the size of the arrays
  
  if(stride < 0)	//We must find a previous state given the current state
  {
      stride = -stride;			//Set stride to the absolute distance
	  dist = stride/rand_deg;
	  mod = stride%rand_deg;
	  iterativeRand(-mod);	//Stride = -(rand_deg*dist + mod). This line handles the mod part 

	  if(dist == 0)		//Bail out early if we are done
	  {
		  errorCode = exportSlibcState();	//Export the state thats being asked to be permuted
		  if(sem_post(&unrankGlobalSem))
			  return UNRANK_FAILED_TO_GET_SEMAPHORE;
		  return errorCode;
	  }

	   //Setup the arrays. To swap arrays we swap pointers
	  errorCode = initStaticArray(&t1,n);		//Allocate three arrays for the job
	  if(errorCode)
	  {		
			sem_post(&unrankGlobalSem);
			return errorCode;	//We are dead at this point
	  }
	  errorCode = initStaticArray(&t2,n); //This code is duplicated so we don't 
	  if(errorCode)
	  {
			deallocStaticArray(&t1);
			sem_post(&unrankGlobalSem);
			return errorCode;	//We are dead at this point
	  }
	  errorCode = initStaticArray(&t3,n);		//allocate before we know we need to
	  if(errorCode)
	  {
			deallocStaticArray(&t1);
			deallocStaticArray(&t2);
			sem_post(&unrankGlobalSem);
			return errorCode;	//We are dead at this point
	  }

	  s_source = t1;
	  result = t2;	
	  tmp = t3;	
	  identity(result,n);

	  createPermMatrixRev(s_source);	//Set up s_source to be equiv to calling rand_reverse()
  }
  else
  {
	  dist = stride/rand_deg;
	  mod = stride%rand_deg;
	  iterativeRand(mod);	//Stride = rand_deg*dist + mod. This line handles the mod part 

	  if(dist == 0)		//Bail early
	  {
		  errorCode = exportSlibcState();	//Export the state thats being asked to be permuted
		  if(sem_post(&unrankGlobalSem))
			  return UNRANK_FAILED_TO_GET_SEMAPHORE;
		  return errorCode;
	  }
	  //Setup the arrays. To swap arrays we swap pointers
	   //Setup the arrays. To swap arrays we swap pointers
	  errorCode = initStaticArray(&t1,n);		//Allocate three arrays for the job
	  if(errorCode) 
	  {
			sem_post(&unrankGlobalSem);
			return errorCode;	//We are dead at this point
	  }
	  errorCode = initStaticArray(&t2,n); //This code is duplicated so we don't 
	  if(errorCode)
	  {
			deallocStaticArray(&t1);
			sem_post(&unrankGlobalSem);
			return errorCode;	//We are dead at this point
	  }
	  errorCode = initStaticArray(&t3,n);		//allocate before we know we need to
	  if(errorCode)
	  {
			deallocStaticArray(&t1);
			deallocStaticArray(&t2);
			sem_post(&unrankGlobalSem);
			return errorCode;	//We are dead at this point
	  }

	  s_source = t1;
	  result = t2;	
	  tmp = t3;	
	  identity(result,n);

	  createPermMatrix(s_source);	//Set up the s_source matrix to be equivalent to a forward
									//iterations
  }

  while(dist)	//Compute a quick matrix power by successive doubling
  {
      if(dist&1)	//If this is odd we include this power of the matrix in our accumulated
					//matrix power

      {
          matrixMult(s_source,result,tmp,n);
          swap = tmp;
          tmp = result;
          result = swap;
      }

      dist = dist>>1;	//Put the next most least-significant-bit in place for next iteration

      if(dist)		//If there is more to do
      {
          matrixMult(s_source,s_source,tmp,n);	//compute the next power
          swap = tmp;			//Swap pointers to avoid data copies
          tmp = s_source;
          s_source = swap;
      }
  }
  updateStateMatrix(result,n);		//The state has now been permuted by the correct number
									//of iterations
  deallocStaticArray(&t1);		//Must not leak memory
  deallocStaticArray(&t2);
  deallocStaticArray(&t3);


  errorCode = exportSlibcState();	//Export the state thats being asked to be permuted
  if(sem_post(&unrankGlobalSem))
	  return UNRANK_FAILED_TO_GET_SEMAPHORE;
  return errorCode;
}

long int t_random_backwords()
/*
PURPOSE: Improved random number generator for you temporally dyslexics out there. 
         This function computes random numbers in the opposite order as random() does
*/
{
  long int tmp;
  
  if (rand_type == TYPE_0)
  {
    tmp = state[0];    
    state[0] = ((state[0]-RAND_B)*invPowerArray[0]) & LONG_MAX;
    //Calculate prev = (cur-b)*a^(-1) mod (LONG_MAX+1)
    
    return tmp;
  }
  else
  {
    rptr--;
    
    if(rptr < state)
    {
	  rptr = end_ptr - 1;
      fptr--;
    }
    else
    {
      fptr--;
      if(fptr < state)
        fptr = end_ptr - 1;
    }
    
    tmp = ((*fptr) >> 1) & LONG_MAX;
    *fptr -= *rptr;   //Undo the previous addition
    
    return tmp;
  }    
}

// ************************ Needed functions ************************
#define N_LONG_BITS   (sizeof(unsigned long)*8)   //Number of bits in a long

const rand_int32_t invPowerArray[31] = { //Array for A = 0x41c64e6d
	0x6eb9eb65,        // A^(-1)
	0x5c6c95d9,        // A^(-2)
	0x3ece51f1,        // A^(-4)
	0x361664e1,        // A^(-8)
	0x6a6c8dc1,        // A^(-16)
	0x3d562b81,        // A^(-32)
	0x3c109701,        // A^(-64)
	0x31322e01,        // A^(-128)
	0x62a85c01,        // A^(-256)
	0x2660b801,        // A^(-512)
	0x51017001,        // A^(-1024)
	0x3302e001,        // A^(-2048)
	0x2a05c001,        // A^(-4096)
	0x640b8001,        // A^(-8192)
	0x8170001,        // A^(-16384)
	0x102e0001,        // A^(-32768)
	0x205c0001,        // A^(-65536)
	0x40b80001,        // A^(-131072)
	0x1700001,        // A^(-262144)
	0x2e00001,        // A^(-524288)
	0x5c00001,        // A^(-1048576)
	0xb800001,        // A^(-2097152)
	0x17000001,        // A^(-4194304)
	0x2e000001,        // A^(-8388608)
	0x5c000001,        // A^(-16777216)
	0x38000001,        // A^(-33554432)
	0x70000001,        // A^(-67108864)
	0x60000001,        // A^(-134217728)
	0x40000001,        // A^(-268435456)
	0x1,        // A^(-536870912)
	0x1         // A^(-1073741824)
};

rand_int32_t invPowerOfA(rand_int32_t power)
/*
PURPOSE: Calculates A^(-power).
*/
{
  rand_int32_t retVal;
  int i;
  
  retVal = 1;
  i = 0;
  power &= 0xFFFFFFFF;    //Just incase this is a 64 bit machine...
  
  while(power)
  {
    if(power & 1)
        retVal = (retVal*invPowerArray[i]) & LONG_MAX;
    i++;
    power = power>>1;
  }
  
  return retVal;    
}

unsigned long fastPoly(unsigned long a, unsigned long n)
/*
PURPOSE: This function computes (a^n + a^(n-1) ... + 1)mod (LONG_MAX+1) in log(n) time
*/
{
    int i;
    int x;
    unsigned long powers[N_LONG_BITS];
    unsigned long tmp;
    
    
    x = log_2(n);
    powers[0]=1;
    powers[1]=a;
    for(i=2;i<=x;i++)    //Preliminary work
    {
      //Powers is filled so that powers[x] = a^n, powers[x-1] = a^(n/2) ... powers[1] = a,
      //and powers[0] = 0. This is essentially a fast power function computed in the 
      //reverse order as is typical with intermediate results saved.
      
      powers[i] = (powers[i-1]*powers[i-1]) & LONG_MAX;
      if (x-i)  //Isolate the i most significant bits of n
          tmp = n >> (x-i);   //the most i significant bits of n
      else
          tmp = n;
      if(tmp & 1) //If this is odd then we need an extra multiple of A
          powers[i] = (powers[i]*powers[1]) & LONG_MAX;      
    }

    return fastPolyItr(powers,x,n);
}

unsigned long fastPolyItr(unsigned long* powers, unsigned long x, unsigned long n)
/*
PURPOSE: Helper function for fastPoly(). It does the main work of the algorithm, which works
by a divide and concur routine.

NOTE: I have no idea if anyone else has ever done this. 

NOTE: x is relates to n in that x = log_2(n). This is used to compute intermediate results
*/
{
    unsigned long tmp;
        
   //Default cases
    if(x == 0)      //Poly is a^0 = 1
        return 1;
    if(x == 1)      //Poly is a + 1
        return powers[1] + 1;
    
    if( n & 1)
    {
        tmp = (fastPolyItr(powers,x-1,n>>1)*(powers[x-1]*powers[1]+1)) & LONG_MAX;
        //    (a^(n/2) + ...     + a + 1  )*(a^(n/2 + 1) + 1)         mod (LONG_MAX+1) 

    }
    else    //if n is even
    {
        tmp = (fastPolyItr(powers,x-1,n>>1)*(powers[x-1]+1)) & LONG_MAX;
        //    (a^(n/2) + ...     + a + 1  )*(a^(n/2) + 1)   mod (LONG_MAX+1)
        //tmp now equals a^n....+1 + a^(n/2) so we subtract out the extra factor of a^(n/2)
        tmp = (tmp - powers[x-1]) & LONG_MAX;    
    }
    return tmp;       
}

int log_2(unsigned long n)
/*
PURPOSE: This function computes floor(log_base_2(n))
         This function is a helper function of fastPoly()
*/
{
  int i;
  
  i=0;
  while(n)
  {
      n = n>>1;
      i++;
  } 
  
  return i;
}
unsigned long fastPower(unsigned long a, unsigned long n)
/*
PURPOSE: This function computed a^n mod (LONG_MAX+1). 
		 (LONG_MAX+1) is equal to 2147483647, which happens to be prime.
*/
{
	unsigned long tmp;
	unsigned long result;

	a = a & LONG_MAX;
	tmp = a;
	result = 1;

	while(n)	//Standard fast power algorithm
	{
		if(n & 1)
			result = (result*tmp) & LONG_MAX;

		tmp = (tmp*tmp) & LONG_MAX;
		n = n>>1;
	}

	return result;
}

void matrixMult(rand_int32_t** left,rand_int32_t** right,
				rand_int32_t** res,int n)
/*
PURPOSE: This function multiplies left*right where left and right are two matrices 
         of size n*n. This is a helper function to unrankRand().
         
         The result is placed within the res matrix which must also be of size n*n.
*/
{
  int i,j;
  rand_int32_t *rowLeft;
  rand_int32_t *colRight;
  rand_int32_t *rowEnd;

    for(i=0;i<n;i++)
    {
        for(j=0;j<n;j++)
        {
            res[i][j] = 0;
            rowLeft = left[i];
            rowEnd = rowLeft + n;
            colRight = right[0] + j;
            for(;rowLeft < rowEnd; rowLeft++,colRight+= n)
                res[i][j] += (*rowLeft)*(*colRight);
        }
    }  
/*
    int i,j,k;

    for(i=0;i<n;i++)
    {
        for(j=0;j<n;j++)
        {
            res[i][j] = 0;
            for(k=0;k<n;k++)
            {
                res[i][j] += left[i][k]*right[k][j];
            }
        }
    }   
*/
}

int initStaticArray(rand_int32_t*** p,int n)
/*
PURPOSE: This function allocates an array for the unrank function. The format is there is
         one unsigned long* array containing n pointers. These n pointers point into an
         n*n sized allocation of longs for a total of two allocations.
         This is compatible with c's a[i][j] calling convention for array indexing
*/
{
    int i;
    
    rand_int32_t* allocation;

    *p = (rand_int32_t**) malloc(sizeof(rand_int32_t*)*n);  //Allocate the backbone pointer
	if(!(*p))
		return UNRANK_RAND_FAILED_MEMORY_ALLOC;
    allocation = (rand_int32_t*) malloc(sizeof(rand_int32_t)*n*n); //Allocate the area for data
	if(!allocation)
	{
		free((void*) (*p));           //Free the backbone
		*p = 0;

		return UNRANK_RAND_FAILED_MEMORY_ALLOC;
	}
    
    for(i=0; i<n ;i++)    //Place the data in non-overlapping segments
        (*p)[i] = allocation + (i*n);

	return 0;
}

void deallocStaticArray(rand_int32_t*** p)
/*
PURPOSE: Free an array allocated with initStaticArray()
*/
{
   free((void*) ((*p)[0]) );     //Free the main allocation
   free((void*) (*p));           //Free the backbone
   *p = 0;                       //Easier to debug memory leaks if you zero out old pointers
}

void updateStateMatrix(rand_int32_t ** a, int n)
/*
PURPOSE: This code uses the matrix a to update the current state array. This function
         Assumes that a[][] has been set up to permute the current state to the current
         state. This is essentially multiplies a matrix be a vector
*/
{
    int i,j;

    rand_int32_t tmp[LARGEST_POLY_SIZE];

    for(i=0;i<n;i++)    //Take the ith row
    {
        tmp[i] = 0;
        for(j=0;j<n;j++)//Do the dot product with the state vector
            tmp[i]+=a[i][j]*state[j]; //The result is saved in a temp vector
    }

    for(i=0;i<n;i++) //Move the temp variable back to the state array
        state[i] = tmp[i];

}

void createPermMatrix(rand_int32_t** a)
/*
PURPOSE: This function creates a matrix that will permute the current state equivalently
         to calling random rand_deg times. This is the core of what makes unrankRand() 
         work. 
*/
{
    int n;
    int i,k;
    int lag,start,rear;
    int offset;

    n = rand_deg;           //Set up some information about the polynomial
    lag = rand_sep;
    offset = fptr - state;

    identity(a,n);      //Start with the identity matrix
    for(i=0;i<n;i++)    //Duplicate the action of rand_deg calls to random() by identically
                        //permuting the matrix. This takes advantage of the fact that order
                        //does not matter for multiplication/division in modular arithmetic 
                        //(i.e. its abealian).
    {
        start = (i+offset)% n;    //Set the pointers up to know which rows to sum       
        if(start < lag)       
            rear = n-(lag-start);
        else
            rear = start-lag;
        
        for(k=0;k<n;k++)    //Sum the start row and the end row into the start row duplicating      
        	a[start][k]+=a[rear][k];  //The action of the ith call to random()
    }

}

void createPermMatrixRev(rand_int32_t** a)
/*
PURPOSE: This function creates a matrix that will permute the current state equivalently
         to calling random_backwords() rand_deg times. This function allows unrankRand() to
		 take negative numbers as parameters.
*/
{
    int n;
    int i,k;
    int lag,start,rear;
    int offset;

    n = rand_deg;           //Set up some information about the polynomial
    lag = rand_sep;
    offset = fptr - state;

    identity(a,n);      //Start with the identity matrix

    for(i=0;i<n;i++)    //Invert the action of rand_deg calls to random() by identically
                        //permuting the matrix. This takes advantage of the fact that order
                        //does not matter for multiplication/division in modular arithmetic 
                        //(i.e. its abealian). 
    {
        start = (n-i-1+offset)% n;    //Start the inversion with the previous row and work       

        if(start < lag)				  //backwards.
            rear = n-(lag-start);
        else
            rear = start-lag;
        
        for(k=0;k<n;k++)    //Sub the start row and the end row into the start row duplicating
		{
            a[start][k]-=a[rear][k];  //The action of the ith call to random_backword()
		}

    }
}

void iterativeRand(long stride)
/* PURPOSE: This function finds the i'th random state from the current state 
   iteratively. The function works with negative or positive numbers
*/
{
	if( stride > 0 )
		while(stride--)
			t_random();
	else
		while(stride++)
			t_random_backwords();
}

int importSlibcState()
/* PURPOSE: This function copies the state of glibc's random function and
   places it into state information local to this file. Returns non-null if
   there was an error.
   
   NOTICE: This relies on internal information of glibc's random state format
   if the format that glibc changes then this function and everything based on
   it (basically all of unrankRand()) will break as well.
*/
{
	rand_int32_t tmpArray[2];	//Temporarily place information on the stack
	rand_int32_t * curState;
	int type;
	int rear;
	int i;

	tmpArray[0] = TYPE_0;	//Just fake it for a couple clock cycles
	curState = (rand_int32_t *) setstate ((PTR)tmpArray);

	type = curState[0] % MAX_TYPES;
	rear = curState[0] / MAX_TYPES;

	switch (type)
    {
    case TYPE_0:
    case TYPE_1:
    case TYPE_2:
    case TYPE_3:
    case TYPE_4:
      rand_type = type;
      rand_deg = degrees[type];
      rand_sep = seps[type];
      break;
    default:
      // State info munged.  
      errno = EINVAL;
      return -1;
    }
	if(rand_type != TYPE_0)
		for(i=0;i<rand_deg;i++)	//Copy the state array
			state[i] = curState[i+1];
	else
		state[0] = curState[1];

	if (rand_type != TYPE_0)
    {
      rptr = &state[rear];
      fptr = &state[(rear + rand_sep) % rand_deg];
    }
	// Set end_ptr too.  
	end_ptr = &state[rand_deg];

	setstate((PTR)curState);	//Restore the old state information

	return 0;
}

int exportSlibcState()
/*
 Move the internal state to random()'s state
*/
{
	rand_int32_t tmpArray[2];	//Temporarily place information on the stack
	rand_int32_t* curState;
	int i;

	tmpArray[0] = TYPE_0;	//Just fake it for a couple clock cycles
	curState = (rand_int32_t*) setstate ((PTR)tmpArray);

	if (rand_type == TYPE_0)	//Wrap up the pointer information
		curState[0] = rand_type;
	else
		curState[0] = (MAX_TYPES * (rptr - state)) + rand_type;

	if(rand_type != 0)
		for(i=0;i<rand_deg;i++)	//Copy the state array
			curState[i+1] = state[i];
	else 
		curState[1] = state[0];
	//This could be very fatal if they have changed the size between 
	//import and export //TODO Check this

	setstate((PTR)curState);	//Restore the old state information

	return 0;
}

void identity(rand_int32_t** a,int n)
/*
PURPOSE: Initialize the matrix a to the identity matrix assume the matrix is n*n
*/
{
    int i,j;

    for(i=0;i<n;i++)
    {
      for(j=0;j<n;j++)
      {
        if(i==j)
          a[i][j]=1;
        else
          a[i][j]=0;
      }
    }
}

// **************************** Parallel interface functions *****************************
int* createProcessRandomTable(int me, int numProc, unsigned long n,
                            unsigned long* myRange)
/*
PURPOSE:    This function creates a table of random values for each process. This action
            is usefull for testing programs like parallel search routines or parallel
            sorting algorithms, where a set of integers is distributed accross a cluster.
			All processes are started off with a TYPE_3 random number generator intialized
			initialized consitanly with a serial program that starts by calling:
			srandom(RANDOM_SEED);

PARAMITERS:
 me:        The process ID of the process calling this function (ID range from 0 to numProc-1)
 numProc:   The number of processes being used in the parallelization (often called p)
 n:         The number of random integers to generate. This many numbers are distributed
              evenly across the cluster with each process recieving about n/numSalves of them
 myRange:   A pointer where the function will store how many integers the caller was allocated
              A process will either be allocated n/myRange or n/myRange+1 numbers so that the 
              total range across the cluster is equal to n. If myRange is 0 it is ignored.

RETURN VALUE:
 The function returns an array of size myRange where random integers have been stored. At
 this point the process has been initialized with a subsequence and is ready for work. The
 function returns NULL on an error. Check the errno value against the error codes listed later
 in this file if this occurs. Note for very small problem sizes it may occur that some
 process has no work to do. In this case myRange is set to 0 and the function returns a
 invalid but non-null pointer (equal to -1).
*/
{
	unsigned long start, range, i;
	unsigned long rem;
	int* array = 0;
	char* oldState;
	char* newState;
	int errorCode;

	if(myRange)
		*myRange = 0;

	if((numProc <= 0)||(me < 0)||(me >= numProc))//Some wise guy....
	{
		errno = INVALID_NUMBER_OF_PROCS;
		return 0;
	}
	
	rem = n%numProc;		//Does the number of processes divide the range
	start = me*(n/numProc);	//Where we would start if rem == 0...
	if( me <= rem )		//Account for the first rem states which have one extra int
		start += me; 
	else
		start += rem;

	range = (n/numProc);
	if( me < rem )			//Increment the range for the first rem states
		range++;

	//printf("Process %d starting at %ld and ending at %ld (not including) using %ld\n",me,start,start+range,range);

	if(!range)	//we can exit if we don't have any work. random() is set to a consistent place.
		return (int*) -1;
	
	newState = (char*) malloc(BREAK_3);		//Set up random() regardless of if we have any work
	if(!newState)
	{
		errno = UNRANK_RAND_FAILED_MEMORY_ALLOC;
		return 0;
	}
	oldState = initstate(RANDOM_SEED, newState, BREAK_3);	

	if(start)	//If unrankRand() is needed
	{
		errorCode = unrankRand(start);	//Something happened so we bail
		if( errorCode )
		{
			setstate(oldState);		//Restore the old random state
			free((void*) newState);	//Free what we allocated
			errno = errorCode;
			return 0;
		}
	}

	array = (int*) malloc(range*sizeof(int));
	if(!array)	//We died on memory....
	{
		setstate(oldState);		//Restore the old random state
		free((void*) newState);	//Free what we allocated
		errno = UNRANK_RAND_FAILED_MEMORY_ALLOC;
		return 0;
	}

	if(myRange)		//Report the length of the allocation
		*myRange = range;

	for(i=0;i<range;i++)		//seed the array
		array[i] = random();

	setstate(oldState);		//Restore the old random state
	free((void*) newState);	//Free what we allocated

	return array;
}


int seedParallel(unsigned long seed, int me, unsigned long long int numsPerProc)
/*
PURPOSE: This function initializes a processes random number generator. Each process will be
         set up so that it can call random() numsPerProc times before it's stream of 
         random numbers coincides with another processes. 

         To use this function each process would have to know:
            1) Its own process number (where IDs start at 0 and run up continuously to p-1)
            2) How many numbers will be consumed by each processor. For this
               function to work all processes must call random() an equal number of times
            3) A random seed that is the same value at every process. This can be a
                pound define or can be part of the information that is sent to each process

PARAMETERS:
 seed: A random seed given to all processes
 me:   An int that represents this process. This function assumes that
       the process ID start at 0 and increase contiguously
 numsPerProc:   The number of random numbers that will be consumed by one
                process. This function uses this to tell where to initialize
                each process.

RETURN VALUE:
 The function returns false if the action failed. Error codes are listed farther down
 in this file.
*/
{
	unsigned long long total;

	if( me < 0 )	//Bad Smeagol. Very bad. Trying to trick poor innocent hobbit-hackers... 
		return INVALID_NUMBER_OF_PROCS;

	srandom(seed);	//For this function we trust the user to only call initstate if
					//they know what they are doing...

	total = me*numsPerProc;
	if(total)
		return unrankRand(total);

	return 0;
}

long long int seedParallelTotal(unsigned long seed, 
                                          int me,
                                          int p,
                                long long int itr,
                                long long int grain)
/*
PURPOSE: This function is like seedParallel() except that it splits 
         the number range for the user.

PARAMETERS:
itr: The total number of loop iterations. The total number of random's
     consumed will be itr*grain
me: My process ID where process ID's range from 0 to p-1
p:  number of processes
seed: a random seed that must be equal on all processes
grain: Granularity of the usage of random numbers. When using block
       parallelization, the number of random numbers used per
       main loop iteration must be taken into account. For instance
       if 2 randoms are used per loop iteration then it makes no
       sense to give a process an odd number of random numbers. grain
       allows the function to know how to dole out work.

RETURN VALUE:
The function returns the number of loop iterations to perform.
The function returns a negative number upon an error and sets the
errno value appropriately.
*/ 
{
	long long start, range;
	long long rem;
	int errorCode;

	if((me < 0)||(p<0)||(me>=p))
	{
		errno = INVALID_NUMBER_OF_PROCS;
		return -1;
	}

	srandom(seed);	//For this function we trust the user to only call initstate if
					//they know what they are doing...

	if((!grain)||(!itr))	//We have no work to do
		return 0;

	rem = itr%p;		//Does the number of processes divide the range
	start = me*(itr/p);	//Where we would start if rem == 0...
	if( me <= rem )		//Account for the first rem states which have one extra int
		start += me; 
	else
		start += rem;
	start *= grain;		//We need grain many randoms per loop iteration

	range = (itr/p);
	if( me < rem )			//Increment the range for the first rem states
		range++;

	if(start)	//If we have work to do
	{
		errorCode  = unrankRand(start);
		if(errorCode)
		{
			errno = errorCode;
			return -1;	//Signal an error
		}
	}

	return range;	//Return the total number of loop iterations to perform
}

// **************************** Advanced interface ************************

struct randomObject* par_generateRandomObject(int type,unsigned long seed)
/* PURPOSE: This function creates a random object that can be used to generate
   a stream of random numbers. The object can then be used when multiple sources
   of random numbers are needed. This makes it simpler to get deterministic resultssem_
   when multiple threads are consuming random numbers. The function returns 0 when
   there is an error and sets the errno value appropriately
*/
{
	struct randomObject* ro;

	if((type > TYPE_4)||(type < 0))
	{
		errno = UNRANK_UNRECOGNIZED_RAND_TYPE;
		return 0;
	}

	ro = (struct randomObject*) malloc(sizeof(struct randomObject));
	if(!ro)	
	{
		errno = UNRANK_RAND_FAILED_MEMORY_ALLOC;
		return 0;
	}
	ro->ready = FALSE;		//Unrank has not been called on this object
	ro->rand_type = type;
	ro->rand_deg = degrees[type];
	ro->rand_sep = seps[type];
	
	ro->state = (long int*) malloc(breaks[type]);
	if(!ro->state)	//I really am sick of checking memory...
	{
		free((void*)ro);
		errno = UNRANK_RAND_FAILED_MEMORY_ALLOC;
		return 0;
	}
	state = state - 1;	//Stupid thing we have to do to support glibc
	ro->end_ptr = &(ro->state[ro->rand_deg]);

	if( sem_init(&(ro->sem),1,1) )	//create a semaphore
	{
		free((void*)ro->state + 1);		//Give it all away
		free((void*)ro);
		errno = UNRANK_FAILED_TO_GET_SEMAPHORE;
		return 0;
	}

    par_srandom(seed,ro);	//Seed the resulting random object
			//Don't worry about error code because I know the object is valid

	return ro;
}

int par_freeRandomObject(struct randomObject* ro)
/* PURPOSE: Free the memory associated with a ro 
   This function returns non-null if there was an error
*/
{
	if(!ro)
		return UNRANK_BAD_POINTER;
	if(ro->state)	//Should be there but check
		free((void*) (ro->state - 1) );
	if(sem_destroy(&(ro->sem)) )
			return UNRANK_FAILED_TO_GET_SEMAPHORE;
	free((void*) ro);

	return 0;
}

int par_srandom(unsigned int seed, struct randomObject* ro)
/* PURPOSE: Thread safe version of srandom() see man pages. The function returns
   FALSE if the operation completed successfully and fails otherwise
*/
{
//This function derives from a similar function in gnu's glibc code
    long int i;
	int errorCode;

	if(!ro)		//Move random()s state
	{
		errorCode = par_getSem(0);	//Get the master semaphore
		if(errorCode)
			return errorCode;

		srandom(seed);	//We now own random()s state..

		errorCode = par_releaseSem(0);	//Release the master semaphore
		if(errorCode)
			return errorCode;

		return 0;
	}

	//ELSE we are modifying an object

	errorCode = par_getSem(ro);		//Take the semaphore
	if(errorCode)
		return errorCode;

	ro->state[0] = seed;
	if (ro->rand_type != TYPE_0)		//This is glibc code essentially
    {
      for (i = 1; i < ro->rand_deg; ++i)
			ro->state[i] = (1103515145 * ro->state[i - 1]) + 12345;
      ro->fptr = &(ro->state)[ro->rand_sep];
      ro->rptr = &(ro->state)[0];
      for (i = 0; i < 10 * ro->rand_deg; ++i)
		 par_random_raw(ro);	//We own the semaphore so we must call an internal function
    }

	errorCode = par_releaseSem(ro);	//Release the semaphore
	if(errorCode)
		return errorCode;

	return 0;
}

long int par_random_raw(struct randomObject* ro)
/* WARNING: This function offers no protection. If you use it with a bad pointer
   your code will die.

  This function is like glibs random() function except that it takes a ro object
  reference. The raw refers to the fact that it ignores the semaphore lock on the
  object. Use at your own risk.
*/
{
// This function derives it's implementation from random.c...

	if(!ro)		//Give glibs random state
		return random();

	if (ro->rand_type == TYPE_0)
	{
	  ro->state[0] = ((ro->state[0] * 1103515245) + 12345) & LONG_MAX;
	  return ro->state[0];
	}
	else
	{
		long int i;

		*(ro->fptr) += *(ro->rptr);
		/* Chucking least random bit.  */
		i = (*(ro->fptr) >> 1) & LONG_MAX;
		++(ro->fptr);
		if (ro->fptr >= ro->end_ptr)
		{
			ro->fptr = ro->state;
			++(ro->rptr);
		}
		else
		{
			++(ro->rptr);
			if (ro->rptr >= ro->end_ptr)
				ro->rptr = ro->state;
		}
		return i;
   }
}

int par_getSem(struct randomObject* ro)
/* PURPOSE: Take the semaphore from the randomObject. If ro is 0 then get the master
   semaphore. This function returns non-null if there was a problem.

NOTE: Please call par_initialize() before calling this function with ro=0. Failing
      to do so can lead to a miss synchronization, meaning that the state of random()
	  will be inconsistent with atomic random operations.
*/
{
	int errorVal;

    if(!ro)
	{
		if( masterSem_init == FALSE)	//Didn't initialize yet
		{
			errorVal = par_initialize();
			if( errorVal )	//create the master semaphore
				return errorVal;
		}
		if( sem_wait(&masterSem) )		//Retrieve it.
			return UNRANK_FAILED_TO_GET_SEMAPHORE;
		return 0;
	}

	//Else use the object
/*	if( ro->sem == INVALID_SEM_VALUE)	//Didn't initialize yet
	{
		if( sem_init(&(ro->sem),1,1) )	//create a semaphore
			return UNRANK_FAILED_TO_GET_SEMAPHORE;
	}
*/
	if( sem_wait(&(ro->sem)) )		//Retrieve it.
		return UNRANK_FAILED_TO_GET_SEMAPHORE;

	return 0;
}
int par_releaseSem(struct randomObject* ro)
/*
PURPOSE: Release a semaphore object associated with ro. The master semaphore is released
 if 0 is passed in. The semaphore must have been initialized first before this function
 can be called. the easiest way to mess this up is to call par_releaseSem(0);
 without calling par_initialize();
*/
{
	if(!ro)
	{

		if( masterSem_init == FALSE)	//Didn't initialize yet!!
			return UNRANK_ASKED_TO_RELEASE_A_NONE_EXISTANT_SEMAPHORE;
		if( sem_post(&masterSem) )		//Return it.
			return UNRANK_FAILED_TO_GET_SEMAPHORE;

		return 0;
	}

	//Else use the object
/*
	if( ro->sem == INVALID_SEM_VALUE)	//Didn't initialize yet
		return UNRANK_ASKED_TO_RELEASE_A_NONE_EXISTANT_SEMAPHORE;
*/	if( sem_post(&(ro->sem)) )		//Return it.
		return UNRANK_FAILED_TO_GET_SEMAPHORE;

	return 0;
}

int par_initialize()
/* PURPOSE: Basically just initialize the masterSem semaphore. This semaphore protects the
   state of random(). The function returns non-null if there was an error.
*/
{
	if( masterSem_init == FALSE)	//Didn't initialize yet
	{
		if( sem_init(&masterSem,1,1) )	//create a semaphore
			return UNRANK_FAILED_TO_GET_SEMAPHORE;
		masterSem_init = TRUE;
	}
	if( unrankGlobalSem_init == FALSE)	//Didn't initialize yet
	{
		if( sem_init(&unrankGlobalSem,1,1) )	//create a semaphore
			return UNRANK_FAILED_TO_GET_SEMAPHORE;
		unrankGlobalSem_init = TRUE;
	}

	return 0;
}

long int wait_random(struct randomObject* ro)
/*
PURPOSE: This function work under the premise that most people will call unrankRand()
once on a given stream and then let any slaves have at it. This function will hold off
the slaves until the object has been made ready by the master calling unrankRand() on
the random stream.
*/
{
	if(!ro)
	{
		while(!masterReady)
			sleep(1);
		return random();
	}
	while(!ro->ready)
		sleep(1);
	return par_random_raw(ro);
}

/*
char* par_wrapupRandomObject(struct randomObject* ro)
//PURPOSE: create an array that can be used in a setstate() call to 
//   initialize random(). This is basically a copy function.
//
{
	int errorCode;
	char* newState;

	if(!ro)
	{

		return (char*) -1;
	}

	errorCode = par_getSem(ro);		//Take the semaphore
	if(errorCode)
	{
		errno = errorCode;
		return 0;
	}



	newState = (char*) malloc(breaks[ro->rand_type]);	//Allocate enough space
	if(!newState)	//We died on memory allocation
	{
		par_releaseSem(ro);	//Release the semaphore	(don't bother with error code
		errno = UNRANK_RAND_FAILED_MEMORY_ALLOC;
		return 0;
	}

	errorCode = par_releaseSem(ro);	//Release the semaphore
	if(errorCode)
	{
		errno = errorCode;
		return 0;
	}

	return par_setstate(newState,0);
}
*/

//char* par_exportRandomObject(struct randomObject* ro)
/* PURPOSE: This function sets random()'s state to match that of ro. It
   returns 0, and sets errno if there was a problem. It returns -1 if
   you try to export the masters state to the master
*/

